import asyncdispatch, jester, json, jwt, logging, options, sequtils, strutils,
  times, timeutils, uuids
import ./db, ./configuration, ./models, ./service, ./version

const JSON = "application/json"

type
  Session* = object
    user*: User
    issuedAt*, expires*: Time

proc newSession*(user: User): Session =
  result = Session(
    user: user,
    issuedAt: getTime().utc.trimNanoSec.toTime,
    expires: daysForward(1).trimNanoSec.toTime)

template halt(code: HttpCode,
              headers: RawHeaders,
              content: string): typed =
  ## Immediately replies with the specified request. This means any further
  ## code will not be executed after calling this template in the current
  ## route.
  bind TCActionSend, newHttpHeaders
  result[0] = CallbackAction.TCActionSend
  result[1] = code
  result[2] = some(headers)
  result[3] = content
  result.matched = true
  break allRoutes

template jsonResp(code: HttpCode, details: string = "", headers: RawHeaders = @{:} ) =
  halt(
    code,
    headers & @{"Content-Type": JSON},
    $(%* {
      "statusCode": code.int,
      "status": $code,
      "details": details
    })
  )

template json500Resp(ex: ref Exception, details: string = ""): void =
  when not defined(release): debug ex.getStackTrace()
  error details & ":\n" & ex.msg
  jsonResp(Http500)

proc toJWT*(ctx: PMApiContext, session: Session): string =
  ## Make a JST token for this session.
  var jwt = JWT(
    header: JOSEHeader(alg: HS256, typ: "jwt"),
    claims: toClaims(%*{
      "sub": $(session.user.id),
      "iat": session.issuedAt.toUnix.int,
      "exp": session.expires.toUnix.int }))

  jwt.sign(ctx.cfg.authSecret)
  result = $jwt

proc fromJWT*(ctx: PMApiContext, strTok: string): Session =
  ## Validate a given JWT and extract the session data.
  let jwt = toJWT(strTok)
  var secret = ctx.cfg.authSecret
  if not jwt.verify(secret): raiseEx "Unable to verify auth token."
  jwt.verifyTimeClaims()

  # Find the user record (if authenticated)
  let userId = jwt.claims["sub"].node.str
  var user: User
  try: user = ctx.db.getUser(parseUUID(userId))
  except: raiseEx "unknown user"

  result = Session(
    user: user,
    issuedAt: fromUnix(jwt.claims["iat"].node.num),
    expires: fromUnix(jwt.claims["exp"].node.num))

proc extractSession(ctx: PMApiContext, request: Request): Session =
  ## Helper to extract a session from a reqest.

  # Find the auth header
  if not request.headers.hasKey("Authorization"):
    raiseEx "No auth token."

  # Read and verify the JWT token
  let headerVal = request.headers["Authorization"]
  if not headerVal.startsWith("Bearer "):
    raiseEx "Invalid Authentication type (only 'Bearer' is supported)."

  result = fromJWT(ctx, headerVal[7..^1])

proc makeAuthToken*(ctx: PMApiContext, email, pwd: string): string =
  ## Given a user's email and pwd, validate the combination and generate a JWT
  ## token string.

  if email.len == 0 or pwd.len == 0:
    raiseEx "fields 'username' and 'password' required"

  # find the user record
  var user: User
  try:
    let users = ctx.db.getUserByEmail(email)
    if users.len != 1: raiseEx ""
    user = users[0]
  except: raiseEx "invalid username or password"

  if not validatePwd(user, pwd): raiseEx "invalid username or password"

  let session = newSession(user)

  result = toJWT(ctx, session)

template checkAuth() =
  ## Check this request for authentication and authorization information.
  ## Injects the session into the running context. If the request is not
  ## authorized, this template returns an appropriate 401 response.

  var session {.inject.}: Session

  try: session = extractSession(ctx, request)
  except:
    debug "Auth failed: " & getCurrentExceptionMsg()
    jsonResp(Http401, "Unauthorized", @{"WWW-Authenticate": "Bearer"})

proc start*(ctx: PMApiContext): void =

  if ctx.cfg.debug: setLogFilter(lvlDebug)

  var stopFuture = newFuture[void]()

  settings:
    port = Port(ctx.cfg.port)
    appName = "/api"

  routes:

    get "/version":
      resp($(%("personal_measure_api v" & PM_API_VERSION)), JSON)

    post "/auth-token":
      var email, pwd: string
      try:
        let jsonBody = parseJson(request.body)
        email = jsonBody["email"].getStr
        pwd = jsonBody["password"].getStr
      except: jsonResp(Http400)

      try:
        let authToken = makeAuthToken(ctx, email, pwd)
        resp($(%authToken), JSON)
      except: jsonResp(Http401, getCurrentExceptionMsg())

    get "/user":
      checkAuth()

      resp(Http200, $(%session.user), JSON)

    post "/service/debug/stop":
      if not ctx.cfg.debug: jsonResp(Http404)
      else:
        let shutdownFut = sleepAsync(100)
        shutdownFut.callback = proc(): void = complete(stopFuture)
        resp($(%"shutting down"), JSON)

    get "/api-tokens":
      checkAuth()

      resp(Http200, $(%ctx.db.getApiTokenByUserId($session.user.id)))

    post "/api-tokens":
      checkAuth()

      var newToken: ApiToken
      try:
        let jsonBody = parseJson(request.body)
        newToken = ApiToken(
          id: genUUID(),
          userId: session.user.id,
          name: jsonBody["name"].getStr,
          expires: none[DateTime](),
          hashedToken:  "")
      except: jsonResp(Http400)

      try:
        let tokenValue = "" # TODO
        newToken.hashedToken = hashPwd(tokenValue)
        ctx.db.createApiToken(token)
        let respToken = %newToken
        newToken["value"] = tokenValue
        resp($newToken, JSON)

  waitFor(stopFuture)

